/***************************************************************************
***
***    Copyright 2005  Hon Hai Precision Ind. Co. Ltd.
***    All Rights Reserved.
***    No portions of this material shall be reproduced in any form without the
***    written permission of Hon Hai Precision Ind. Co. Ltd.
***
***    All information contained in this document is Hon Hai Precision Ind.  
***    Co. Ltd. company private, proprietary, and trade secret property and 
***    are protected by international intellectual property laws and treaties.
***
****************************************************************************/
 
#include "lib_types.h"
#include "lib_string.h"
#include "lib_queue.h"
#include "lib_malloc.h"
#include "lib_printf.h"
#include "cfe.h"
#include "cfe_iocb.h"
#include "cfe_devfuncs.h"
#include "cfe_error.h"
#include "cfe_fileops.h"
#include "cfe_loader.h"
#include "ui_command.h"
#include "bsp_config.h"

#include <typedefs.h>
#include <osl.h>
#include <sbutils.h>
#include <sbconfig.h>
#include <sbchipc.h>
#include <sbmemc.h>
#include <bcmendian.h>
#include <bcmnvram.h>
#include <bcmutils.h>
#include <trxhdr.h>
#include "addrspace.h"
#include "initdata.h"

#include "bsp_priv.h"

#include "board_config.h"

#include "tftpd.h"


static int isHexDigit(char c)
{
    if (c >= '0' && c <= '9')
        return TRUE;
    else
    if (c >= 'A' && c <= 'F')
        return TRUE;
    else
    if (c >= 'a' && c <= 'f')
        return TRUE;
    else
        return FALSE;
}

static int isValidMacAddr(char *macAddr)
{
    int i, j;
    
    for (i=0, j=0; i<6; i++)
    {
        if (!isHexDigit(macAddr[j++]))
            return FALSE;
       
        if (!isHexDigit(macAddr[j++]))
            return FALSE;

        if (i != 5 && macAddr[j++] != ':')
            return FALSE;

        if (i == 5 && macAddr[j] != '\0')
            return FALSE;
    }

    return TRUE;
}

extern int bcm_robo_check_link_status(void);
static unsigned long calculate_checksum (int action, char *s, int size)
{
    static unsigned long c0, c1;
    unsigned long checksum, b;
    int i;

    switch (action)
    {
        case 0:
            c0 = c1 = 0;
            break;

        case 1:
            for (i = 0; i < size; i++)
            {
                c0 += s[i] & 0xff;
                c1 += c0;
                
                if ((i % 90000) == 0)
                {
                    printf(".");
                    bcm_robo_check_link_status();
                }
            }
            break;
    
        case 2:
            b = (c0 & 65535) + ((c0 >> 16) & 65535);
            c0 = ((b >> 16) + b) & 65535;

            b = (c1 & 65535) + ((c1 >> 16) & 65535);
            c1 = ((b >> 16) + b) & 65535;

            checksum = ((c1 << 16) | c0);

            return checksum;
    }
    return 0;
}

int verify_checksum(char *buf, unsigned long buf_len, unsigned long chksum)
{
    unsigned long cal_chksum = 0;

    printf("Loading .");
    calculate_checksum (0, NULL, 0);
    calculate_checksum (1, buf, buf_len);
    cal_chksum = calculate_checksum (2, NULL, 0); 
    printf("\n");
    

    if (cal_chksum != chksum) {
//        printf("Checksum mismatch: \n");
        printf("Image chksum: 0x%08X\n", chksum);
        printf("Calc  chksum: 0x%08X\n", cal_chksum);
        return -1;
    }

#ifdef _DEBUG
    printf("verify_checksum calculate_checksum    = 0x%08X\n", cal_chksum);
    printf("verify_checksum image_chksum          = 0x%08X\n", chksum);
#endif

    return 0;   
}

int verify_board_id(char *buf)
{
    unsigned char *board_id = (unsigned char *)BOARD_DATA_ADDR;

    if (strcmp(buf, board_id) != 0)
    {
        /* if it cannot read the board id from the board_data section, try to read it from nvram. */
        if ( nvram_match("boardid", buf) == 0 )
        {
            /* printf("Board ID mismatch!\n"); */
            printf("Board ID : %s\n", board_id);
            printf("Image ID : %s\n", buf);
            return -1;
        }
    }

    return 0;
}

int read_board_data(void)
{
    struct board_data *bd = (struct board_data *)BOARD_DATA_ADDR;
    char mac[6];
    char buf[30];
    /*char boot_wait[2];*/

    /* Read LAN MAC */
    memcpy(mac, bd->lan_mac, 6);
    sprintf(buf, "%02x:%02x:%02x:%02x:%02x:%02x",
            (unsigned char)mac[0],
            (unsigned char)mac[1],
            (unsigned char)mac[2],
            (unsigned char)mac[3],
            (unsigned char)mac[4],
            (unsigned char)mac[5]);
            
    if ( isValidMacAddr(buf) && strcmp(buf, "00:00:00:00:00:00") )
        nvram_set("et0macaddr", buf);
    else
        nvram_set("et0macaddr", "00:00:00:00:00:01");

    /* Read WAN MAC */
    memcpy(mac, bd->wan_mac, 6);
    sprintf(buf, "%02x:%02x:%02x:%02x:%02x:%02x",
            (unsigned char)mac[0],
            (unsigned char)mac[1],
            (unsigned char)mac[2],
            (unsigned char)mac[3],
            (unsigned char)mac[4],
            (unsigned char)mac[5]);
            
    if ( isValidMacAddr(buf) && strcmp(buf, "00:00:00:00:00:00"))
        nvram_set("et1macaddr", buf);
    else
        nvram_set("et1macaddr", "00:00:00:00:00:01");

    return 0;
}

static int gpio_control(int gpio, int value)
{
    sb_gpioouten(sbh, 1 << gpio, 1 << gpio, GPIO_HI_PRIORITY);
    sb_gpioout(sbh, 1 << gpio, value << gpio, GPIO_HI_PRIORITY);

    return 0;
}

int test_led_control(int on_off)
{
    int tftpd_state = get_tftpd_state();

/*
    if (tftpd_state == TFTPD_STATE_OFF)
        gpio_control(GPIO_LED_TEST , 1);
    else if ( (tftpd_state == TFTPD_STATE_WAIT_IMAGE) || (tftpd_state == TFTPD_STATE_WRITE_IMAGE) )
        gpio_control(GPIO_LED_TEST , on_off);
*/

    /* if it need to use power led instead of test led in wgr614v8 board rev. L1*/
    /* modified start by EricHuang, 11/28/2006 */
    if (tftpd_state == TFTPD_STATE_OFF)
    {
        sb_gpioouten(sbh, ((unsigned long) 1 << 7), ((unsigned long) 1 << 7), GPIO_HI_PRIORITY);
        sb_gpioout(sbh, (unsigned long)1 << 7, (unsigned long)0 << 7, GPIO_HI_PRIORITY);
    }
    else if ( (tftpd_state == TFTPD_STATE_WAIT_IMAGE) || (tftpd_state == TFTPD_STATE_WRITE_IMAGE) )
    {
        sb_gpioouten(sbh, ((unsigned long) 1 << 7), ((unsigned long) 1 << 7), GPIO_HI_PRIORITY);
        if (on_off==1)
            sb_gpioout(sbh, (unsigned long)1 << 7, (unsigned long)0 << 7, GPIO_HI_PRIORITY);
        else
            sb_gpioout(sbh, (unsigned long)1 << 7, (unsigned long)1 << 7, GPIO_HI_PRIORITY);
    }
    /* modified end by EricHuang, 11/28/2006 */

    return 0;
}

int test_led_toggle(void)
{
    static int led_state = 0;
    if (led_state == 0)
        led_state = 1;
    else
        led_state = 0;

    test_led_control(led_state);
    return 0;
}
/*
int wlan_led_control(int on_off)
{
    gpio_control(GPIO_LED_WLAN , on_off);
    return 0;
}
*/
int wan_led_control(int on_off)
{
    gpio_control(GPIO_LED_WAN , on_off);
    return 0;
}

static char empty_buf[0x10000];

int check_reset(void)
{
    int val = sb_gpioin(sbh);

    if (!(val & (1 << GPIO_RESET)))
    {
        int i;
        int nvram_fh;

        /* Reset pin is low (active).
         * Wait 3 sec to see if we need to load default
         */
        for (i=0; i<30; i++)
        {
            cfe_sleep(CFE_HZ/10);
            
            /* disable the output bit first for reading gpio 7 */
            sb_gpioout(sbh, (unsigned long)1 << GPIO_RESET, (unsigned long)1 << GPIO_RESET, GPIO_HI_PRIORITY); /* pull high first for old hardware revision */
            sb_gpioouten(sbh, ((unsigned long) 1 << GPIO_RESET), ((unsigned long) 0 << GPIO_RESET), GPIO_HI_PRIORITY);

            val = sb_gpioin(sbh);
            
            /* enable the output bit to turn on the power led */
            enable_power_led();

            if (val & (1 << GPIO_RESET))
                return 0;
        }

        /* Change Power LED to green */
//        gpio_control(GPIO_LED_GREEN , 1);
//        gpio_control(GPIO_LED_AMBER , 0);
        
        printf("load default!\n");

        nvram_fh = cfe_open("flash1.nvram");
        memset(empty_buf, 0xFF, sizeof(empty_buf));
        cfe_writeblk(nvram_fh, 0, empty_buf, sizeof(empty_buf));
        cfe_close(nvram_fh);
        
        cfe_sleep(CFE_HZ);

        ui_docommand("reboot");

    }

    return 0;
}

int lan_led_control(int port, int status)
{
    gpio_control(port , status);
    return 0;
}

void enable_power_led(void)
{
    /* enable the output bit to turn on the power led */
    sb_gpioouten(sbh, ((unsigned long) 1 << GPIO_RESET), ((unsigned long) 1 << GPIO_RESET), GPIO_HI_PRIORITY);
    sb_gpioout(sbh, (unsigned long)1 << GPIO_RESET, (unsigned long)0 << GPIO_RESET, GPIO_HI_PRIORITY);    
}
